import java.util.*;

import org.apache.commons.collections.map.AbstractReferenceMap;
import org.apache.commons.collections.map.ReferenceIdentityMap;
import org.apache.commons.collections.set.MapBackedSet;

public aspect ASyncContainsAll {
	
	dependency {
		strong sync(c), starAll(c,*);
	}	

	dependency {
		strong sync(c), starAll(*,c);
	}	
	
	Set synched = MapBackedSet.decorate(new ReferenceIdentityMap(AbstractReferenceMap.WEAK,AbstractReferenceMap.HARD,true),new Object()); 
	
	dependent synchronized after sync() returning(Collection c):
		!within(dacapo..*) &&
		!within(ASyncContainsAll) &&
		call(* Collections.synch*(..)) {
		synched.add(c);
	}
	
	dependent synchronized Object around starAll(Collection t, Collection a):
		!within(dacapo..*) &&
		!within(ASyncContainsAll) &&
		call(* Collection.*All(Collection)) && target(t) && args(a) {
		if(t!=a &&	//we don't invoke c.containsAll(c) (in this case we have the lock on c already)
		   synched.contains(t) && synched.contains(a) &&	//t and a were both synchronized
		   !Thread.holdsLock(a)) {	//we don't have the lock on a
			//give error
//			System.err.println("Must synchronize argument collection at: " +
//					thisJoinPointStaticPart.getSourceLocation());
			//actually *do* synchronize and proceed
			synchronized(a) {
				return proceed(t,a);
			}
		}
		//in any other case, just proceed as usual
		return proceed(t,a);
	}	
	
	before(): call(* dacapo.Benchmark.run(..)) {
		//reset all data structures for new benchmark run
		synched.clear();
	}

}

//class ASyncContainsAllTest {
//	
//	public static void main(String[] args) {
//		List l1 = Collections.synchronizedList(new ArrayList());
//		List l2 = Collections.synchronizedList(new ArrayList());
//		l1.addAll(l2);
//	}
//	
//}