import java.util.*;

import org.apache.commons.collections.map.AbstractReferenceMap;
import org.apache.commons.collections.map.ReferenceIdentityMap;
import org.apache.commons.collections.set.MapBackedSet;

public aspect ASyncIterationCollection {
	
	dependency {
		strong sync, createIter;		
	}
	dependency {
		strong sync, createIter, iter;		
	}
	
    Set synched = MapBackedSet.decorate(new ReferenceIdentityMap(AbstractReferenceMap.WEAK,AbstractReferenceMap.HARD,true),new Object());
    Map iteratorToColl = new ReferenceIdentityMap(AbstractReferenceMap.WEAK,AbstractReferenceMap.WEAK,true);
	
    dependent synchronized after sync() returning(Collection c):
		!within(dacapo..*) &&
    	!within(ASyncIterationCollection) &&
    	call(* Collections.synchr*(..)) {
    	synched.add(c);
    }
    
    dependent synchronized after createIter(Collection c) returning(Iterator i):
		!within(dacapo..*) &&
    	!within(ASyncIterationCollection) &&
    	call(* Collection+.iterator()) && target(c) {
    	if(synched.contains(c)) {
	    	if(!Thread.holdsLock(c)) {
	   		 	//System.err.println("MATCHED at "+thisJoinPointStaticPart.getSourceLocation());
	    	} else {
	    		iteratorToColl.put(i,c);
	    	}
    	}
    }
    
    dependent synchronized before iter(Iterator i):
		!within(dacapo..*) &&
    	!within(ASyncIterationCollection) &&
    	call(* Iterator.*(..)) && target(i) {
    	Collection c = (Collection) iteratorToColl.get(i);
    	if(c!=null && !Thread.holdsLock(c)) {
   		 	//System.err.println("MATCHED at "+thisJoinPointStaticPart.getSourceLocation());
    	}
    }
    
	before(): call(* dacapo.Benchmark.run(..)) {
		//reset all data structures for new benchmark run
		synched.clear();
		iteratorToColl.clear();
	}
}
//
//class ASyncIterationCollectionTest {
//	
//	public static void main(String[] args) {
//		List l = Collections.synchronizedList(new ArrayList());
//		Iterator i;
//		synchronized(l) {
//			i = l.iterator();
//		}
//		i.hasNext();
//		i = l.iterator();
//	}
//	
//}
