import java.util.*;

import org.apache.commons.collections.map.AbstractReferenceMap;
import org.apache.commons.collections.map.ReferenceIdentityMap;
import org.apache.commons.collections.set.MapBackedSet;

public aspect ASyncIterationMap {
	
	dependency {
		strong sync, createColl, createIter;		
	}
	dependency {
		strong sync, createColl, createIter, iter;		
	}
	
    Set synched = MapBackedSet.decorate(new ReferenceIdentityMap(AbstractReferenceMap.WEAK,AbstractReferenceMap.HARD,true),new Object());
    Map iteratorToColl = new ReferenceIdentityMap(AbstractReferenceMap.WEAK,AbstractReferenceMap.WEAK,true);
    Map collToMap = new ReferenceIdentityMap(AbstractReferenceMap.WEAK,AbstractReferenceMap.WEAK,true);
	
    dependent synchronized after sync() returning(Map m):
       !within(dacapo..*) &&
 	   !within(ASyncIterationMap) &&
    	call(* Collections.synchr*(..)) {
    	synched.add(m);
    }
    
    dependent synchronized after createColl(Map m) returning(Collection c) :
	   !within(dacapo..*) &&
	   !within(ASyncIterationMap) &&
	   (call(* Map.keySet()) || call(* Map.values()) || call(* Map.entrySet())) && target(m) {
    	if(synched.contains(m))
    		collToMap.put(c,m);
    }
    
    dependent synchronized after createIter(Collection c) returning(Iterator i):
 	   !within(dacapo..*) &&
 	   !within(ASyncIterationMap) &&
    	call(* Collection+.iterator()) && target(c) {
    	Map m = (Map)collToMap.get(c);
    	if(m!=null && synched.contains(m)) {
	    	if(!Thread.holdsLock(m)) {
	   		 	//System.err.println("MATCHED at "+thisJoinPointStaticPart.getSourceLocation());
	    	}
	    	iteratorToColl.put(i,c);
    	}
    }
    
    dependent synchronized before iter(Iterator i):
 	   !within(dacapo..*) &&
 	   !within(ASyncIterationMap) &&
    	call(* Iterator.*(..)) && target(i) {
    	Collection c = (Collection) iteratorToColl.get(i);
    	if(c!=null) {
        	Map m = (Map)collToMap.get(c);
    		if(m!=null && !Thread.holdsLock(m)) {
   		 		//System.err.println("MATCHED at "+thisJoinPointStaticPart.getSourceLocation());
    		}
    	}
    }
    
	before(): call(* dacapo.Benchmark.run(..)) {
		//reset all data structures for new benchmark run
		synched.clear();
		iteratorToColl.clear();
		collToMap.clear();
	}
    
}

//class ASyncIterationMapTest {
//	
//	public static void main(String[] args) {
//		Map m = new HashMap();
//		Map sm = Collections.synchronizedMap(m);
//		Set km = m.keySet();
//		Set ksm = sm.keySet();
//		km.iterator();	//needs no sync.
//		ksm.iterator();	//not synchronized -> match!
//		Iterator i;
//		synchronized(sm) {
//			i = ksm.iterator(); //correctly synchronized			
//		}
//		synchronized(ksm) {
//			ksm=null;
//			km=null;
//			m=null;
//			sm=null;
//	     	System.gc();
//	     	System.runFinalization();
//	     	System.gc();
//	     	System.gc();
//			i.hasNext();	//incorrectly synchronized -> match!
//		}
//	}
//	
//}
