import java.util.*;

import org.apache.commons.collections.map.AbstractReferenceMap;
import org.apache.commons.collections.map.ReferenceIdentityMap;
import org.apache.commons.collections.set.MapBackedSet;

public aspect FailSafeEnum {

	dependency {
		strong createEnum, update, next;
	}
	
     pointcut vector_update() :
    	call(* Vector.add*(..)) ||
        call(* Vector.clear()) ||
        call(* Vector.insertElementAt(..)) ||
        call(* Vector.remove*(..)) ||
        call(* Vector.retainAll(..)) ||
        call(* Vector.set*(..));
     
     Map vecToEnums = new ReferenceIdentityMap(AbstractReferenceMap.WEAK,AbstractReferenceMap.WEAK,true);
     Set tainted = MapBackedSet.decorate(new ReferenceIdentityMap(AbstractReferenceMap.WEAK,AbstractReferenceMap.HARD,true),new Object());

    dependent synchronized after createEnum(Vector v) returning(Enumeration e) :
 	     !within(dacapo..*) &&
    	 !within(FailSafeEnum) && 
    	 ((call(* Vector+.elements()) && target(v)) || (call(Enumeration+.new(..)) && args(v))) {
    	 cleanup();
    	 Set enums = (Set)vecToEnums.get(v);
    	 if(enums==null) {
    		 enums = MapBackedSet.decorate(new ReferenceIdentityMap(AbstractReferenceMap.WEAK,AbstractReferenceMap.WEAK,true),new Object());
    		 vecToEnums.put(v,enums);
    	 }
    	 enums.add(e);
    }
     
    dependent synchronized before next(Enumeration e):
	    !within(dacapo..*) &&
    	!within(FailSafeEnum) &&
     	call(Object Enumeration.nextElement()) && target(e) {
    	 cleanup();
    	 if(tainted.contains(e)) {
    		 //System.err.println("MATCHED at "+thisJoinPointStaticPart.getSourceLocation());
    	 }     	  	 
     }
     
    dependent synchronized after update(Vector v) returning:
	     !within(dacapo..*) &&
    	 !within(FailSafeEnum) &&
    	 vector_update() && target(v) {
    	 cleanup();
    	 Set enums = (Set)vecToEnums.get(v);
    	 if(enums!=null) {
        	 tainted.addAll(enums);
    	 }
    }
     
     void cleanup() {
    	 for (Iterator iter = vecToEnums.values().iterator(); iter.hasNext();) {
    		 Set s = (Set) iter.next();
    		 if(s.isEmpty()) {
    			 iter.remove();
    		 }
		 }
     }
     
 	before(): call(* dacapo.Benchmark.run(..)) {
		//reset all data structures for new benchmark run
 		vecToEnums.clear();
		tainted.clear();
	}
     
}

//class FailSafeEnumTest {
//    public static void main(String[] args) {
//    	Vector v = new Vector();
//    	v.add("1");
//    	v.add("2");
//    	Enumeration e = v.elements();
//    	e.nextElement();
//    	v.add("3");
//    	v=null;
//     	System.gc();
//     	System.runFinalization();
//     	System.gc();
//     	System.gc();
//    	e.nextElement();
//	}
//}
