import java.util.*;

import org.apache.commons.collections.map.AbstractReferenceMap;
import org.apache.commons.collections.map.ReferenceIdentityMap;
import org.apache.commons.collections.set.MapBackedSet;

public aspect FailSafeEnumHashtable {

	dependency {
		strong createEnum, update, next;
	}
	
   pointcut hashtable_update() :
    	call(* Map.put*(..)) ||
        call(* Map.putAll*(..)) ||
        call(* Map.clear()) ||
        call(* Map.remove*(..));

	   Set toMonitor = MapBackedSet.decorate(new ReferenceIdentityMap(AbstractReferenceMap.WEAK,AbstractReferenceMap.HARD,true),new Object());
	   Set changed = MapBackedSet.decorate(new ReferenceIdentityMap(AbstractReferenceMap.WEAK,AbstractReferenceMap.HARD,true),new Object());
	   Map enumToMap = new ReferenceIdentityMap(AbstractReferenceMap.WEAK,AbstractReferenceMap.WEAK,true);
   
	   
	   dependent synchronized after createEnum(Hashtable h) returning(Enumeration e) :
	 	   !within(dacapo..*) &&
		   !within(FailSafeEnumHashtable) && 
		   (call(* Hashtable+.elements()) || call(* Hashtable+.keys())) && target(h) {
	    	 cleanup();
	    	 toMonitor.add(h);
	    	 enumToMap.put(e,h);	    	
	   }
	    
	   dependent synchronized before next(Enumeration e):
	 	   !within(dacapo..*) &&
		   !within(FailSafeEnumHashtable) && 
		   call(Object Enumeration.nextElement()) && target(e) {
	    	 cleanup();
	    	 Hashtable h = (Hashtable)enumToMap.get(e);
	    	 if(h!=null && changed.contains(h)) {
	    		 //System.err.println("MATCHED at "+thisJoinPointStaticPart.getSourceLocation());
	    	 }  
	   }
        
	   dependent synchronized after update(Hashtable h) returning:
	 	   !within(dacapo..*) &&
		   !within(FailSafeEnumHashtable) && 
		   hashtable_update() && target(h) {
	    	 cleanup();
	    	 if(toMonitor.contains(h)) {
	    		 changed.add(h);
	    	 }
	   }
	   
	   //remove all vectors for which all enumerations were GC'ed
	   void cleanup() {
		   for (Iterator iter = changed.iterator(); iter.hasNext();) {
			   Hashtable h = (Hashtable) iter.next();
			   if(!enumToMap.containsValue(h)) {
				   	toMonitor.remove(h);
			   }
		   }
		   changed.retainAll(toMonitor);
	   }
	   
	   before(): call(* dacapo.Benchmark.run(..)) {
			//reset all data structures for new benchmark run
		   toMonitor.clear();
		   changed.clear();
		   enumToMap.clear();
	   }	     
}

//class FailSafeEnumHashtableTest {
//    public static void main(String[] args) {
//    	Hashtable h = new Hashtable();
//    	h.put("1","2");
//    	h.put("2","3");
//    	Enumeration e = h.keys();
//    	e.nextElement();
//    	h.put("3","4");
//    	h=null;
//     	System.gc();
//     	System.runFinalization();
//     	System.gc();
//     	System.gc();
//    	e.nextElement();
//	}
//}