import java.util.*;

import org.apache.commons.collections.map.AbstractReferenceMap;
import org.apache.commons.collections.map.ReferenceIdentityMap;
import org.apache.commons.collections.set.MapBackedSet;

public aspect FailSafeIter {

	dependency {
		strong createIter, update, next;
	}
	
    pointcut collection_update(Collection c) :
        (  call(* java.util.Collection+.add*(..)) ||
           call(* java.util.Collection+.clear()) ||
           call(* java.util.Collection+.retainAll(..)) ||
           call(* java.util.Collection+.remove*(..)) ) && target(c);
     
     Map collToIterators = new ReferenceIdentityMap(AbstractReferenceMap.WEAK,AbstractReferenceMap.WEAK,true);
     Set tainted = MapBackedSet.decorate(new ReferenceIdentityMap(AbstractReferenceMap.WEAK,AbstractReferenceMap.HARD,true),new Object());

     dependent synchronized after createIter(Collection c) returning(Iterator i) :
	 	 !within(dacapo..*) &&
    	 !within(FailSafeIter) &&
    	 ((call(* Collection+.iterator()) && target(c)) || (call(Iterator+.new(..)) && args(c))) {
    	 Set its = (Set)collToIterators.get(c);
    	 if(its==null) {
    		 its = MapBackedSet.decorate(new ReferenceIdentityMap(AbstractReferenceMap.WEAK,AbstractReferenceMap.WEAK,true),new Object());
    		 collToIterators.put(c,its);
    	 }
    	 its.add(i);
     }
     
     dependent synchronized before next(Iterator i):
	 	 !within(dacapo..*) &&
    	 !within(FailSafeIter) &&
    	 call(Object Iterator.next()) && target(i) {
    	 if(tainted.contains(i)) {
    		 //System.err.println("MATCHED at "+thisJoinPointStaticPart.getSourceLocation());
    	 }     	 
     }
     
     dependent synchronized before update(Collection c) :
	 	 !within(dacapo..*) &&
    	 !within(FailSafeIter) &&
    	 collection_update(c) {
    	 Set iters = (Set)collToIterators.get(c);
    	 if(iters!=null) {
    		 tainted.addAll(iters);
    		 //clean up in case something got GC'ed
    		 if(iters.isEmpty()) {
    			 collToIterators.remove(c);
    		 }
    	 }
	 }
     
	 before(): call(* dacapo.Benchmark.run(..)) {
	    //reset all data structures for new benchmark run
		collToIterators.clear();
		tainted.clear();
	 }	
     
}

//class FailSafeIterTest { 
//    public static void main(String[] args) {
//     	List c = new ArrayList();
//     	c.add("");
//     	c.add("");
//     	Iterator i = c.iterator();
//     	i.next();
//     	c.add("");
//     	c=null;
//     	System.gc();
//     	System.runFinalization();
//     	System.gc();
//     	System.gc();
//     	i.next();
// 	}
//}

