import java.util.*;

import org.apache.commons.collections.map.AbstractReferenceMap;
import org.apache.commons.collections.map.ReferenceIdentityMap;
import org.apache.commons.collections.set.MapBackedSet;

public aspect FailSafeIterMap {

	dependency {
		strong createColl, createIter, update, next;
	}
	
   pointcut map_update() :
    	call(* Map.put*(..)) ||
        call(* Map.putAll*(..)) ||
        call(* Map.clear()) ||
        call(* Map.remove*(..));

	   Map mapToColls = new ReferenceIdentityMap(AbstractReferenceMap.WEAK,AbstractReferenceMap.WEAK,true);
	   Map collToIterators = new ReferenceIdentityMap(AbstractReferenceMap.WEAK,AbstractReferenceMap.WEAK,true);
	   Set tainted = MapBackedSet.decorate(new ReferenceIdentityMap(AbstractReferenceMap.WEAK,AbstractReferenceMap.HARD,true),new Object());
	   
	   dependent synchronized after createColl(Map m) returning(Collection c) :
		   !within(dacapo..*) &&
		   !within(FailSafeIterMap) &&
		   (call(* Map.keySet()) || call(* Map.values()) || call(* Map.entrySet())) && target(m) {
	    	 Set colls = (Set)mapToColls.get(m);
	    	 if(colls==null) {
	    		 colls = MapBackedSet.decorate(new ReferenceIdentityMap(AbstractReferenceMap.WEAK,AbstractReferenceMap.HARD,true),new Object());
	    		 mapToColls.put(m,colls);
	    	 }
	    	 colls.add(c);
	   }

	   dependent synchronized after createIter(Collection c) returning(Iterator i) :
		   !within(dacapo..*) &&
		   !within(FailSafeIterMap) &&
		   call(* Collection.iterator()) && target(c) {
	    	 Set its = (Set)collToIterators.get(c);
	    	 if(its==null) {
	    		 its = MapBackedSet.decorate(new ReferenceIdentityMap(AbstractReferenceMap.WEAK,AbstractReferenceMap.HARD,true),new Object());
	    		 collToIterators.put(c,its);
	    	 }
	    	 its.add(i);
	   }

	   dependent synchronized before next(Iterator i):
		   !within(dacapo..*) &&
		   !within(FailSafeIterMap) &&
		   call(* Iterator.next()) && target(i) {
	    	 if(tainted.contains(i)) {
	    		 //System.err.println("MATCHED at "+thisJoinPointStaticPart.getSourceLocation());
	    	 }  
	   }
        
	   dependent synchronized after update(Map m) returning:
		   !within(dacapo..*) &&
		   !within(FailSafeIterMap) &&
		   map_update() && target(m) {
	    	 Set colls = (Set)mapToColls.get(m);
	    	 if(colls!=null) {
		    	 //clean up, if possible
		    	 if(colls.isEmpty())
		    		 mapToColls.remove(m);
	    		 for (Iterator iter = colls.iterator(); iter.hasNext();) {
					Collection coll = (Collection) iter.next();
					Set iters = (Set)collToIterators.get(coll);
					if(iters!=null) {
				    	 //clean up, if possible
				    	 if(iters.isEmpty())
				    		 collToIterators.remove(coll);
						tainted.addAll(iters);
					}
				 } 	    		 
	    	 }
	   }
	   	     
	   before(): call(* dacapo.Benchmark.run(..)) {
	       //reset all data structures for new benchmark run
		   collToIterators.clear();
		   tainted.clear();
		   mapToColls.clear();
	   }	
}

//class FailSafeIterMapTest {
//	public static void main(String[] args) {
//		Map h = new HashMap();
//		h.put("1","2");
//		h.put("2","3");
//		Collection v = h.values();
//		Iterator i = v.iterator();
//		i.next();
//		h.put("3","4");
//		h=null;
//		System.gc();
//		System.runFinalization();
//		System.gc();
//		System.gc();
//		i.next();
//	}
//}
