import java.io.*;
import java.util.*;

import org.apache.commons.collections.map.AbstractReferenceMap;
import org.apache.commons.collections.map.ReferenceIdentityMap;
import org.apache.commons.collections.set.MapBackedSet;

public aspect ReaderTest {
	
		dependency {
			strong create, closeI, readI;
		}
		dependency {
			strong create, closeI, readR;
		}
		dependency {
			strong create, closeR, readI;
		}
		dependency {
			strong create, closeR, readR;
		}	
	
		pointcut createReaderFromIs(InputStream is):
			call(Reader+.new(InputStream+,..)) && args(is,..) ;
		pointcut createReaderFromReader(Reader is):
			call(Reader+.new(Reader+,..)) && args(is,..) ;
		
		//it should be ok to use a weak mapping here because if the reader is strongly referenced then
		//so is the input stream (by the reader) and therefore it will be accessible in the body
		Map readerToInputStream = new ReferenceIdentityMap(AbstractReferenceMap.WEAK,AbstractReferenceMap.WEAK,true);
		//here we have to use strong references because we have to hold on to the readers for each given input stream
		//even if the readers would otherwise be gc'ed
		Map inputStreamToReaders = new HashMap();
		
		Set closedReaders = new HashSet();
		Set closedInputStreams = new HashSet();
		
		dependent synchronized after create(Object is) returning(Reader r):
		    !within(dacapo..*) &&
			!within(ReaderTest) &&
			call(Reader+.new(..)) && args(is,..) && (args(Reader,..)||args(InputStream,..)) {
			readerToInputStream.put(r,is);
			Set set = (Set)inputStreamToReaders.get(is);
			if(set==null) {
				set = MapBackedSet.decorate(new ReferenceIdentityMap(AbstractReferenceMap.WEAK,AbstractReferenceMap.HARD,true),new Object());
				inputStreamToReaders.put(is,set);
			}
			set.add(r);	
		}
		
		dependent synchronized after closeR(Reader r) returning:
		    !within(dacapo..*) &&
			!within(ReaderTest) &&
			call(* close()) && target(r) {
			if(readerToInputStream.containsKey(r)) {
				closedReaders.add(r);
			}
		}
		
		dependent synchronized after closeI(Object is) returning:
		    !within(dacapo..*) &&
			!within(ReaderTest) &&
			call(* close()) && target(is) && (target(Reader)||target(InputStream)) {
			if(inputStreamToReaders.containsKey(is)) {
				closedInputStreams.add(is);
			}
		}
		
		dependent synchronized before readR(Reader r):
		    !within(dacapo..*) &&
			!within(ReaderTest) &&
			(call(* read(..)) || call(* readLine(..))) && target(r) {
			boolean giveError = false;
			if(closedReaders.contains(r)) {
				giveError = true;
			} else {
				Object is = readerToInputStream.get(r);
				if(is!=null) {
					if(closedInputStreams.contains(is)) {
						giveError = true;
					}
				}
			}
			if(giveError) {
				//System.err.println("MATCHED at "+thisJoinPointStaticPart.getSourceLocation());
			}
		}
		
		dependent synchronized before readI(Object is):
		    !within(dacapo..*) &&
			!within(ReaderTest) &&
			(call(* read(..)) || call(* readLine(..))) && target(is) {
			boolean giveError = false;
			if(closedInputStreams.contains(is)) {
				giveError = true;
			} else {
				Set set = (Set)inputStreamToReaders.get(is);
				if(set!=null) {
					for (Iterator iter = set.iterator(); iter.hasNext();) {
						Reader r = (Reader) iter.next();
						if(closedReaders.contains(r)) {
							giveError = true;
						}
					}
				}
			}
			if(giveError) {
				//System.err.println("MATCHED at "+thisJoinPointStaticPart.getSourceLocation());
			}
		}
		
		before(): call(* dacapo.Benchmark.run(..)) {
		    //reset all data structures for new benchmark run
			readerToInputStream.clear();
			inputStreamToReaders.clear();
			closedReaders.clear();
			closedInputStreams.clear();
		}
		
}


//class ReaderTestTest {
//	
//	public static void main(String[] args) throws IOException {
//		{
//			InputStream is = new ByteArrayInputStream(new byte[10]);
//			Reader r = new InputStreamReader(is);
//			is.read();
//			r.close();
//			r = null;
//			System.gc();
//			System.gc();
//			System.runFinalization();
//			System.gc();
//			is.read();
//		}
//		{
//			InputStream is = new ByteArrayInputStream(new byte[10]);
//			Reader r = new InputStreamReader(is);
//			r.read();
//			is.close();
//			r.read();
//		}
//		{
//			InputStream is = new ByteArrayInputStream(new byte[10]);
//			Reader r = new InputStreamReader(is);
//			Reader r2 = new BufferedReader(r);
//			r.close();
//			try {
//				r2.read();
//			} catch(IOException e){}
//		}
//		{
//			InputStream is = new ByteArrayInputStream(new byte[10]);
//			Reader r = new InputStreamReader(is);
//			r.close();
//			try {
//				r.read();
//			} catch(IOException e){}
//		}
//	}
//	
//}
