import java.io.*;
import java.util.*;

import org.apache.commons.collections.map.AbstractReferenceMap;
import org.apache.commons.collections.map.ReferenceIdentityMap;
import org.apache.commons.collections.set.MapBackedSet;

public aspect WriterTest {
	
		dependency {
			strong create, closeO, writeO;
		}
		dependency {
			strong create, closeO, writeW;
		}
		dependency {
			strong create, closeW, writeO;
		}
		dependency {
			strong create, closeW, writeW;
		}	
	
		pointcut createWriterFromIs(InputStream is):
			call(Writer+.new(OutputStream+,..)) && args(is,..) ;
		pointcut createWriterFromWriter(Writer is):
			call(Writer+.new(Writer+,..)) && args(is,..) ;
		
		//it should be ok to use a weak mapping here because if the writer is strongly referenced then
		//so is the output stream (by the writer) and therefore it will be accessible in the body
		Map writerToOutputStream = new ReferenceIdentityMap(AbstractReferenceMap.WEAK,AbstractReferenceMap.WEAK,true);
		//here we have to use strong references because we have to hold on to the writerss for each given output stream
		//even if the readers would otherwise be gc'ed
		Map outputStreamToWriters = new HashMap();
		
		Set closedWriters = new HashSet();
		Set closedOutputStreams = new HashSet();
		
		dependent synchronized after create(Object is) returning(Writer r):
		    !within(dacapo..*) &&
			!within(WriterTest) &&
			call(Writer+.new(..)) && args(is,..) && (args(Writer,..)||args(OutputStream,..)) {
			writerToOutputStream.put(r,is);
			Set set = (Set)outputStreamToWriters.get(is);
			if(set==null) {
				set = MapBackedSet.decorate(new ReferenceIdentityMap(AbstractReferenceMap.WEAK,AbstractReferenceMap.HARD,true),new Object());
				outputStreamToWriters.put(is,set);
			}
			set.add(r);	
		}
		
		dependent synchronized after closeW(Writer w) returning:
		    !within(dacapo..*) &&
			!within(WriterTest) &&
			call(* close()) && target(w) {
			if(writerToOutputStream.containsKey(w)) {
				closedWriters.add(w);
			}
		}
		
		dependent synchronized after closeO(Object is) returning:
		    !within(dacapo..*) &&
			!within(WriterTest) &&
			call(* close()) && target(is) && (target(Writer)||target(OutputStream)) {
			if(outputStreamToWriters.containsKey(is)) {
				closedOutputStreams.add(is);
			}
		}
		
		dependent synchronized before writeW(Writer w):
		    !within(dacapo..*) &&
			!within(WriterTest) &&
			(call(* write*(..))||call(* print*(..))) && target(w) {
			boolean giveError = false;
			if(closedWriters.contains(w)) {
				giveError = true;
			} else {
				Object is = writerToOutputStream.get(w);
				if(is!=null) {
					if(closedOutputStreams.contains(is)) {
						giveError = true;
					}
				}
			}
			if(giveError) {
				//System.err.println("MATCHED at "+thisJoinPointStaticPart.getSourceLocation());
			}
		}
		
		dependent synchronized before writeO(Object os):
		    !within(dacapo..*) &&
			!within(WriterTest) &&
			(call(* write*(..))||call(* print*(..))) && target(os) && (target(Writer)||target(OutputStream)) {
			boolean giveError = false;
			if(closedOutputStreams.contains(os)) {
				giveError = true;
			} else {
				Set set = (Set)outputStreamToWriters.get(os);
				if(set!=null) {
					for (Iterator iter = set.iterator(); iter.hasNext();) {
						Writer r = (Writer) iter.next();
						if(closedWriters.contains(r)) {
							giveError = true;
						}
					}
				}
			}
			if(giveError) {
				//System.err.println("MATCHED at "+thisJoinPointStaticPart.getSourceLocation());
			}
		}
		
		before(): call(* dacapo.Benchmark.run(..)) {
		    //reset all data structures for new benchmark run
			writerToOutputStream.clear();
			outputStreamToWriters.clear();
			closedWriters.clear();
			closedOutputStreams.clear();
		}
		
}


//class WriterTestTest {
//	
//	public static void main(String[] args) throws IOException {
//		{
//			OutputStream os = new ByteArrayOutputStream();
//			Writer w = new OutputStreamWriter(os);
//			os.write(1);
//			w.close();
//			w = null;
//			System.gc();
//			System.gc();
//			System.runFinalization();
//			System.gc();
//			os.write(1);
//		}
//		{
//			OutputStream os = new ByteArrayOutputStream();
//			Writer w = new OutputStreamWriter(os);
//			w.write(1);
//			os.close();
//			w.write(1);
//		}
//		{
//			OutputStream os = new ByteArrayOutputStream();
//			Writer w = new OutputStreamWriter(os);
//			Writer r2 = new BufferedWriter(w);
//			w.close();
//			try {
//				r2.write(1);
//			} catch(IOException e){}
//		}
//		{
//			OutputStream os = new ByteArrayOutputStream();
//			Writer w = new OutputStreamWriter(os);
//			w.close();
//			try {
//				w.write(1);
//			} catch(IOException e){}
//		}
//	}
//	
//}
