import java.io.*;

public aspect WriterTest {

   /*
	* This aspect checks if a writer is created for an output stream,
	* neither the stream nor the writer were closed when writing to either of the two.   
	*/
	
	pointcut createWriter(OutputStream os):
		call(Writer+.new(OutputStream+,..)) && args(os,..);
	
	tracematch(Writer w, OutputStream os) {
		sym newDaCapoRun before: call(* dacapo.Benchmark.run(..));
		
		sym create after returning(w):( createWriter(os)) && !within(dacapo..*);
		
		sym closeW before:( call(* close()) && target(w)) && !within(dacapo..*);
		sym closeO before:( call(* close()) && target(os)) && !within(dacapo..*);
		
		sym writeW before:( (call(* write(..)) || call(* print*(..))) && target(w)) && !within(dacapo..*);
		sym writeO before:( call(* write(..)) && target(os)) && !within(dacapo..*);
		
		create (writeW|writeO)* (closeW|closeO)+ (writeW|writeO) {
            if(System.getProperty("TMTEST_ACTIVE")!=null) {
                System.err.println("##### Tracematch "+getClass().getName()+" matched!");
            }
		}
	}
	
	pointcut createWriterFromWriter(Writer w):
		call(Writer+.new(Writer+,..)) && args(w,..);
	
	tracematch(Writer w1, Writer w2) {
		sym newDaCapoRun before: call(* dacapo.Benchmark.run(..));
		
		sym create after returning(w1):( createWriterFromWriter(w2)) && !within(dacapo..*);
		
		sym closeW1 before:( call(* close()) && target(w1)) && !within(dacapo..*);
		sym closeW2 before:( call(* close()) && target(w2)) && !within(dacapo..*);
		
		sym writeW1 before:( (call(* write(..)) || call(* print*(..))) && target(w1)) && !within(dacapo..*);
		sym writeW2 before:( (call(* write(..)) || call(* print*(..))) && target(w2)) && !within(dacapo..*);
		
		create (writeW1|writeW2)* (closeW1|closeW2)+ (writeW1|writeW2) {
            if(System.getProperty("TMTEST_ACTIVE")!=null) {
                System.err.println("##### Tracematch "+getClass().getName()+" matched!");
            }
		}
	}

}
