import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Arrays;


public class MemoryMonitor {

	private static final int SLEEP_TIME = 1000;
	private static final boolean EXPLICIT_GC = true;
	
	public static void main(String[] args) throws SecurityException, NoSuchMethodException, ClassNotFoundException, IllegalArgumentException, IllegalAccessException, InvocationTargetException {
		
		ArrayList argList = new ArrayList(Arrays.asList(args));
		String mainClassName = (String) argList.remove(0);
		args = (String[])argList.toArray(new String[0]);
		
		Class mainClass = Class.forName(mainClassName);
		Method[] methods = mainClass.getMethods();
		Method mainMethod = null; 
		for (int i = 0; i < methods.length; i++) {
			Method m = methods[i];
			if(m.getName().equals("main")) {
				mainMethod = m;
				break;
			}
		}
		if(mainMethod==null) {
			throw new RuntimeException("No main method found in "+mainClassName);
		}

		final Thread monitor = new Thread(new Monitor());
		monitor.start();
		
		Runtime.getRuntime().addShutdownHook(new Thread() {
		    public void run() { if(!monitor.isInterrupted()) monitor.interrupt(); }
		});
		
		try {
  			mainMethod.invoke(null, new Object[]{args});
                } finally {				
			monitor.interrupt();
                }
	}

	private static class Monitor implements Runnable {

		public void run() {
			Runtime rt = Runtime.getRuntime();
			NumberFormat df = NumberFormat.getNumberInstance();
			df.setMaximumFractionDigits(3);
			double max = 0;
			double sum = 0;
			int count = 0;
			while(!Thread.currentThread().isInterrupted()) {
				if(EXPLICIT_GC)
					System.gc();
				double mb = (rt.totalMemory()-rt.freeMemory()+0.0)/(1024*1024);
				max = Math.max(max, mb);
				sum += mb;
				count++;
				String mbString = df.format(mb);				
				System.err.println(">>>>> Memory consumption (MB): "+mbString);
				try {
					Thread.sleep(SLEEP_TIME);
				} catch (InterruptedException e) {
					//reset interrupted status
					Thread.currentThread().interrupt();
				}
			}
			String maxString = df.format(max);				
			String avgString = df.format(sum/count);				
			System.err.println("============================================");
			System.err.println(">>>>> Peak memory consumption (MB):    "+maxString);
			System.err.println(">>>>> Number of measurements:          "+count);
			System.err.println(">>>>> Average memory consumption (MB): "+avgString);
		}
	}
	
}
